; Z80FALLx.ASM : FALLZONE game for z80asm on ZX80-1k
;
;  z80asm z80fall#.asm -o z80fall#.o	; ZX80 ver
;  z80asm z80fall#.asm -o z80fall#.p	; ZX81 ver
; 
;  Simple testbed game of ZX80 asm, taking z80test1.asm  
;  Simple part demo version..
; -> z80fall1.asm (23.10.2015)
;  Part working version. 
; -> z80fall2.asm (24.10.2015)
;  Imp borders. Display fall rate.
; -> z80fall3.asm (24.10.2015)
;  Imp PROG_END. Imp wait for key release.
; -> z80fall4.asm (24.10.2015) (256 bytes)
;  Imp freezers
; -> z80fall5.asm (24.10.2015) 
; -> z80fall6.asm (25.10.2015) 
;  Start on ZX81 ver..
; -> z80fall7.asm (26.10.2015) 
;  Imp flicker free routine for ZX80 display (CUSTOM_DISPLAY flag)
;  Works, but too fast.
; -> z80fall8.asm (27.10.2015) 
;  Insert waitrelease loop.
; -> z80fall9.asm (28.10.2015) 
;  Imp (ntargs) multitarget - collect 2/3 targs for level..
;  Imp try-count in io_rand
; -> z80fal10.asm (28.10.2015) 
;  Imp ZX81 ver.. (not working)
; -> z80fal11.asm (29.10.2015) 
;  Imp working ZX81 ver..
; -> z80fal12.asm (30.10.2015) 
;  Imp USE_EXTRA for extended version  2k+ RAM
; -> z80fal12.asm (30.10.2015) 
;  Alter rnd fn
; -> z80fal14.asm (31.10.2015) 
; -> z80fal15.asm (1.11.2015) 
;  Simplify/Improve key scan - use "qweasdzxc" keys..
;  Imp "compatible" ZX81 version (use fast/slow modes)
; -> z80fal16.asm (2.11.2015) 
;   
; NOTES:
;  FALLZONE - A GPL2 licensed open source game (C) A.Millett 1981-2022.
;    (The original BASIC version of this game was written as a test for my own home built ZX81 in 1981.)
; 
; FALLZONE Instructions: use keys QWEADZXC to move yourself (*) around. 
;   Mines fall every time you move, avoid these. 
;   Collect 2 out of 3 targets ($) to go to the next level. 
;   Collect Freezers (F) to halt Mines-falling for a few moves.
;   The higher the level, the faster the mines fall.    
;   The level no (1-9,A-Z) is shown on the top line. (this is the "score")
;   The 1k version has a smaller screen, the 2k version uses 32x24.
;  What level can you reach? Have fun!
;
;  Compiled with my favorite small PD z80 assembler (z80asm.c) by Bas Wijnen.
;

USE_ZX80:  equ 1		; Select target comp..
USE_ZX81:  equ 0
USE_EXTRA: equ 0		; Extra features, 2k+ ram
CUSTOM_DISPLAY: equ 1		; custom (flicker-free) display routine - disable to use more-compatible std routines

;---------------------------------------------------------
; USE_ZX80 - Header code, Basic stub, etc
;---------------------------------------------------------

if USE_ZX80

org 0x4000

PROG_END: equ 0x416f			; Program size

if USE_EXTRA			; Extended version for 2K+..
 STACK_TOP: equ 0x47ff		; Stack top 
 VDUX: equ 32
 VDUY: equ 23
else
 STACK_TOP: equ 0x43ff		; Stack top 1K
 VDUX: equ 27			; Limit screen size for 1k
 VDUY: equ 22
endif


NO_TARGETS: equ 3		; No of targets

pVdu:    equ 400Ch		; Ptr to hires screen (D_FILE)
pEnd:    equ 4010h		; Ptr to screen end (DF_END)
zframes: equ 401eh		; Frame counter (lo first, 1/50 sec)	
zseed:   equ 401ch		; Random seed

io_vsync: equ 0xfe		; Vsync switch

CH_NEWLINE: equ 0x76		; New line char
CH_SPACE:   equ 0
CH_PLAYER:  equ 20
CH_TARGET:  equ 13
CH_MINE:    equ 9
CH_EDGE:    equ 61+128
CH_FREEZER: equ 0x2b
CH_ZERO:    equ 28

zx_getch:  equ 0x013f		; Display screen, read a char.
zx_disp2:  equ 0x01ad		; Display part of screen

playpos:   equ 0x4028		; Player pos (ptr)
randseed:  equ 0x402a		; 3 byte seed for io_rand routine (was  db 59,211,1)
nmine:     equ 0x402d		; Mine count
nfreeze:   equ 0x402e		; Freeze count
ntarg:     equ 0x402f		; Targ count

datastart:			; $4000-$4028 : System variables
 db 0xff,0x88,0xfe,0xff,0xfc,0x40,0xe7,0x03
	; System word ptrs at 0x4008..
 ;   VARS     E_LINE	   D_LINE      DF_EA         DF_END
 dw PROG_END, PROG_END+1, PROG_END+3, PROG_END+4, PROG_END+4

			; Assort system vars
 db 0x02,0x00,0x00,0x00,0x00,0x00
 db 0x00,0x00,0xb0,0x07,0x00,0x00,0x7c,0x3b
 db 0x00,0x00,0x00,0x00,0x21,0x17,0xfc,0x40

			; Basic stub:10 RAND USR(16443)
 db 0x00,0x0a,0xef,0x3a,0x38,0x37,0xda,0x1d,0x22,0x20,0x20,0x1f,0xd9,0x76
			; Basic:  999 REM <RET><RET><CODESTART>..   (Double <RET> prevent list of m/code)
 db 0x03,0xe7,0xfe,0x76,0x76


endif		; USE_ZX80

;---------------------------------------------------------
; USE_ZX81 - Header code, Basic stub, etc
;---------------------------------------------------------

if USE_ZX81

org 0x4009


PROG_END: equ 0x41df		; Program size

if USE_EXTRA			; Extended version for 2K+..
 STACK_TOP: equ 0x47ff		; Stack top 2K+
 VDUX: equ 32
 VDUY: equ 23
else
 STACK_TOP: equ 0x43ff		; Stack top 1K
 VDUX: equ 24
 VDUY: equ 22
endif

NO_TARGETS: equ 3		; No of targets

pVdu:    equ 400Ch		; Ptr to hires screen (D_FILE)
pVars:   equ 4010h		; Ptr to basic vars (VARS)
zframes: equ 4034h		; Frame counter (lo first, 1/50 sec)	

io_vsync: equ 0xfd		; Vsync switch

CH_NEWLINE: equ 0x76		; New line char
CH_SPACE:   equ 0
CH_PLAYER:  equ 23
CH_TARGET:  equ 13
CH_MINE:    equ 8
CH_EDGE:    equ 61+128
CH_FREEZER: equ 0x2b
CH_ZERO:    equ 28

zx_getch:  equ 0x013f		; Display screen, read a char.
zx_disp2:  equ 0x02b5		; Display part of screen

playpos:   equ 0x4048		; Player pos (ptr)
randseed:  equ 0x404a		; 3 byte seed for io_rand routine (was  db 59,211,1)
nmine:     equ 0x404d		; Mine count
nfreeze:   equ 0x404e		; Freeze count
ntarg:     equ 0x404f		; Targ count


datastart:			; $4009 - $407D  System variables
  db 0x00,0x0a,0x00
  db 0xc5,0x41,0xc6,0x41,0xde
  db 0x41,0x00,0x00,0xdf,0x41,0xe3,0x41,0x00
  db 0x00,0xe4,0x41,0xe4,0x41,0x00,0x5d,0x40
  db 0x00,0x02,0x00,0x00,0xbf,0xfd,0xff,0x37
  db 0xc5,0x41,0x00,0x00,0x00,0x00,0x00,0x8d
  db 0x0c,0x00,0x00,0x22,0xad,0x00,0x00,0xbc
  db 0x21,0x18,0x40,0x00,0x00,0x00,0x00,0x00
  db 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
  db 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
  db 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
  db 0x00,0x00,0x00,0x76,0x00,0x00,0x00,0x00
  db 0x00,0x00,0x84,0x00,0x00,0x00,0x84,0xa0
  db 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
  db 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
  db 0x00,0x00,0x00,0x00

  db 0x00,0x0a,0x32,0x01,0xea		; Start of basic:  10 REM..  


 ;remlen: equ (PROG_END-$40ad)
 ;db 0x00,0x14,0x00
 ;    D_LINE       DF_CC         VARS
 ;dw PROG_END-26, PROG_END-25, PROG_END-3,
 ;db 0x00,0x00,0x6f,0x41,0x74,0x41,0x00
 ;db 0x00,0x75,0x41,0x75,0x41,0x00,0x5d,0x40,  0x00,0x02,0x00,0x00,0xbf,0xfd,0xff,0x37
 ;db 0x55,0x41,0x00,0x00,0x00,0x00,0x00,0x8d,  0x0c,0x00,0x00,0xb9,0xd2,0x00,0x00,0xbc
 ;db 0x21,0x18,0x40,0x00,0x00,0x00,0x00,0x00,  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
 ;db 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
 ;db 0x00,0x00,0x00,0x76,0x00,0x00,0x00,0x00,  0x00,0x00,0x85,0x00,0x00,0x00,0x84,0xa0
 ;db 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
 ;db 0x00,0x00,0x00,0x00
 ;db 0x00,0x0a,remlen & 255, remlen/256,0xea		; 0xc2,0x00,
 ;db 0xea		; Start of basic:  10 REM..  

endif			; USE_ZX81

;---------------------------------------------------------
; codestart - main code
;---------------------------------------------------------

codestart:			; Called by USR(16443)
  ld sp,STACK_TOP		; Set SP to top of RAM
if CUSTOM_DISPLAY == 0
 if USE_ZX81
  out (0fdh),a		; Set fast mode 
 endif
endif
  
new_game:			; Start a new game
  ld a,2
  ld (nmine),a		; Init mine count

new_level:		; Start a new level
  ld hl,player_start
  ld (playpos),hl 	; Init player pos center of screen

  xor a
  ld (nfreeze),a	; Clr freeze count

  ld hl,nmine
  inc (hl)		; nmine ++; 

  ld hl,dfile_new	; Reuse basic stub code space
  ld (pVdu),hl		; Update D_FILE pointer

  ld a, CH_NEWLINE	; Place CH_NEWLINE at start of screen.
  ld (hl),a
  inc hl
  ld (hl),a
  inc hl
  
  ld c,VDUY		; Y=24 (C)

loopy:			; Outer loop:  for(y=VDUY;y;y--)..
  ld b,VDUX
  ld a,c
  dec a
  dec a
  cp VDUY-2		; Put edge char on top/bot line
  ld a,CH_SPACE
  jr c,loopx		; Carry set, not border
  ld a,(nmine)		; Display mine fall rate (score) in edge
  add a,CH_ZERO+126
  and 0xbf
  ; ld a,CH_EDGE

loopx:			; Inner loop - fill screen line:  for(x=VDUX;x;x--)..
  ld (hl),a		;
  inc hl		; pVdu ++;
  djnz loopx		; X--; if (X>0) goto loopx; (B)

  ld a, CH_NEWLINE	; Add NEWLINE at the end of each line
  ld (hl),a
  inc hl

  dec c				
  jr nz,loopy		; Y--; if (Y>0) goto loopy; (C)

  ld (hl),a		; Add NEWLINE
  ;inc hl
  ;ld (pVars),hl

  
  ld d,NO_TARGETS	; Draw Targets/Freezers
targloop:
  call io_rand
  ld (hl),CH_TARGET
  call io_rand
  ld (hl),CH_FREEZER
  dec d
  jr nz,targloop

  ld a,NO_TARGETS-1
  ld (ntarg),a


playloop:		; Main game play loop
  ld hl,(playpos)	; Draw player char
  ld (hl),CH_PLAYER

if CUSTOM_DISPLAY		

waitrelease:		; Wait for key-up..
  call displayframe_delay	; Flicker-free display

  ld bc,0x00fe		; Scan for any key down
  in a,(c)
  and 0x1f
  xor 0x1f
  jr nz,waitrelease	; Key still down, loop..

keyloop:

  call displayframe_delay	; Flicker-free display

else			; (CUSTOM_DISPLAY == 0) - Use std routines..

if USE_ZX81
  out (0feh),a		; Set slow mode
endif

waitrelease:		; Scan keyboard, wait till key release

  ld bc,0x00fe
  in a,(c)
  and 0x1f
  cp 0x1f
  jr nz,waitrelease

keyloop:

if USE_ZX80
  call zx_getch		; Read keyboard, display..
endif

endif			; (CUSTOM_DISPLAY)


	; Directly read + decode ZX81/ZX80 Keyboard Matrix  

	;  Port____Line____Bit__0____1____2____3____4__
	;  FEFEh  0  (A8)     SHIFT  Z    X    C    V
	;  FDFEh  1  (A9)       A    S    D    F    G
	;  FBFEh  2  (A10)      Q    W    E    R    T
	;  F7FEh  3  (A11)      1    2    3    4    5
	;  EFFEh  4  (A12)      0    9    8    7    6
	;  DFFEh  5  (A13)      P    O    I    U    Y
	;  BFFEh  6  (A14)    ENTER  L    K    J    H
	;  7FFEh  7  (A15)     SPC   .    M    N    B

			; Use keys about 'S' (QWEASDZXC) for compass dir movement
  ld l,0x07		; mask bits 0,1,2 in (L)
  ld de,0xfffd-VDUX	; de= movedir offset (-VDUX-3)
  ld bc,0xfbfe		; Row "qwert"
  in a,(c)		; Read keyboard
  xor l
  and l			; Flip & Mask bits 0,1,2
  jr nz,key3		; Key hit in (A).. (q=1,w=2,e=4..)

  ld de,0xfffe		; movedir offset  (-2)
  ld b,0xfd		; Row "asdfg"
  in a,(c)		; Read keyboard
  xor l
  and l
  jr nz,key3

  ld de,VDUX-1		; movedir offset
  ld b,0xfe		; Row "Szxcv"
  in a,(c)		; Read keyboard
  srl a			; Bring "zxc" to lo bits
  xor l
  and l
  jr z,keyloop		; No key, screen display loop

key3:			; code 1,2,4 returned
  cp 4
  jr nz,key4
  dec a
key4:
  add a,e		; Compass offset dir now in (A)
  ld e,a
  jr nc,key5
  inc d
key5:

  ld hl,(playpos)
  ld (hl),CH_SPACE		; Clr player pos
  add hl,de			; playpos += movedir
  ld a,(hl)

  cp CH_TARGET
  jr nz,skipnl
  ld a,(ntarg)
  dec a				; ntarg--;
  ld (ntarg),a
  jp nz,drawplayer		; Still more targets to collect..
  jp new_level

skipnl:
  cp CH_FREEZER			; Collect freezer?
  jr nz,skipfreeze
  ld a,(nfreeze)
  add a,4			; Freeze for 4 steps
  ld (nfreeze),a
  jr drawplayer

skipfreeze:
  or a
  jp nz,new_game		; Hit mine/edge, end game

drawplayer:
  ld (playpos),hl		; Save new player position

  ld hl,nfreeze
  ld a,(hl)
  or a
  jr z,dropmines		; No freezer, drop mines
  dec (hl)			; Decrement freezer count
  jp playloop

dropmines:			; Drop some random mines..
  ld a,(nmine)
  ld d,a
mineloop:
  call io_rand			; (HL)=rand empty screen location
  ld a,e
  or a
  jr z,minedone			; e=0, rand failed to find empty sqr
  ld (hl),CH_MINE
  dec d
  jr nz,mineloop
minedone:
  jp playloop

;---------------------------------------------------------
; io_rand - gen rand screen pos
;---------------------------------------------------------

		; A simple psuedo random fn. Return empty screen pos in (HL)
io_rand:
  ld e,127		; Count tries
io_rand1:
  ld a,(randseed)
  ld bc,(randseed+1)
  dec bc
  ld (randseed),bc
  rla 
  rla 
  adc a,b
	; ld h,start/256 :ld l,a : adc a,(hl)   ; a = a ^ PEEK (start + a)
  ld (randseed+2),a
	; Random numbers in A, BC. Now convert to screen pos in (HL)
  ld hl,screen_start
if USE_EXTRA			; Extend to cover larger screen
  srl a				; A=A/2, C=rand 0,1
  jr c, io_rand2
  ld hl,screen_start2		; Alt position to cover bottom of screen
io_rand2:
endif
  and 1
  ld b,a
  add hl,bc
  ld a,(hl)			; Rand pos in (HL) 
  dec e
  ret z				; Limit to 256 tries.. (e=0 means fail)
  or a
  jr nz, io_rand1		; If occupied, try again..
  ret

;---------------------------------------------------------
; USE_ZX80 - Terminating code
;---------------------------------------------------------

if USE_ZX80

if CUSTOM_DISPLAY

displayframe_delay:	; Custom flicker-free display routine
  ld b,64
delay1:
  djnz delay1		; Delay loop during vert retrace

displayframe:
  out (io_vsync),a		; Vert sync off
  
  ld hl,dfile_hi	; Display file loc (bit 15 set)
  ld a,$e9		; line timing count (was $ed) (alter to shift horiz screen pos)
  ld b,25		; 24 rows + 1 for top
  call zx_disp2 	; ($01ad): ROM routine to display top/main screen

  ld a,$f0		; line timing count
  dec hl		; HL points to HALT
  inc b			; Set row count to 1
  dec (iy+$23)		; 47 lines in bottom border
  call zx_disp2		; ($01ad): Show bottom part

  ld (iy+$23),30h	; Reset count
  in a,(io_vsync)	; Vert sync on (Comment out if reading keyboard)
  ret


endif			; (CUSTOM_DISPLAY)

dfile_new:	; New loc for dfile - reuse unused Basic stub code space.

screen_start: equ dfile_new + VDUX+2
screen_start2: equ dfile_new + (VDUX * VDUY) - 512

dfile_hi: equ dfile_new + 0x8000		; Display file himem loc (bit 15 set)
		; Player start center pos Constant (dfile_new must be first)
player_start: equ dfile_new+(VDUX*(VDUY/2))-1

		; Remains of basic stub code. Note:- line length must match for code to run.
 
 ds (PROG_END-1-$)	; Pad unused REM space upto PROG_END with zeros
 db 0x76		; Term REM--..  line

 db 0x80		; Var table terminator

endif		; USE_ZX80

;---------------------------------------------------------
; USE_ZX81 - Terminating code
;---------------------------------------------------------

if USE_ZX81

if CUSTOM_DISPLAY

displayframe_delay:	; Custom flicker-free display routine
  ld b,64
delay1:
  djnz delay1		; Delay loop during vert retrace

displayframe:
  out (io_vsync),a		; Vert sync off
  
  ld hl,dfile_hi	; Display file loc (bit 15 set)
  ld a,$e6		; line timing count (was $ed) (alter to shift horiz screen pos)
  ld c,48		; 48 lines in the bottom border.
  ld b,25		; 24 rows + 1 for top
  call zx_disp2 	; ($02b5): ROM routine to display top/main screen

  ld a,$e7		; line timing count
  dec hl		; HL points to HALT
  inc b			; Set row count to 1
  ld c,47		; 47 lines in the bottom border.
  call zx_disp2		; ($02b5): Show bottom part

  ;ld (iy+$23),30h	; Reset count
  in a,(io_vsync)	; Vert sync on (Comment out if reading keyboard)
  ret

endif			; (CUSTOM_DISPLAY)

dfile_new:	; New loc for dfile - reuse unused Basic stub code space.

screen_start: equ dfile_new + VDUX+2
screen_start2: equ dfile_new + (VDUX * VDUY) - 512

dfile_hi: equ dfile_new + 0x8000		; Display file himem loc (bit 15 set)
		; Player start center pos Constant (dfile_new must be first)
player_start: equ dfile_new+(VDUX*(VDUY/2))-1


		; Remains of basic stub code (10 REM ----..) Note:- line length must match for code to run.
  ds PROG_END-45-$	; Pad unused REM space with zeros
  db  0x76		; Term 10 REM--..  line
			; Basic code: 20 RAND USR 16514
  db  0x00,0x14,0x0e,0x00,0xf9,0xd4,0x1d,0x22
  db  0x21,0x1d,0x20,0x7e,0x8f,0x01,0x04,0x00,0x00,0x76
			; Display file data.. (25 newlines)
  db  0x76,0x76,0x76,0x76,0x76,0x76,0x76,0x76
  db  0x76,0x76,0x76,0x76,0x76,0x76,0x76,0x76
  db  0x76,0x76,0x76,0x76,0x76,0x76,0x76,0x76
  db  0x76
			; Var table terminator
  db  0x80


endif		; USE_ZX81

end

